from boa.blockchain.vm.Neo.Runtime import Notify

from .stuff.things import Awesome, MoreAwesome


def Main(a):

    # create an instance
    m1 = MoreAwesome()

    inpt = 12

    result = m1.add_thirteen_to_this(43)

    result = m1.mmmm(67)

    Notify(result)

    return result



from boa.blockchain.vm.Neo.Runtime import Notify
from boa.blockchain.vm.Neo.Storage import GetContext, Get, Put, Delete


class StorageAPI():

    ctx = GetContext()

    thing = 3

    def getitem(self, key):

        return Get(self.ctx, key)

    def putitem(self, key, value):

        Put(self.ctx, key, value)

    def deleteitem(self, key):

        Delete(self.ctx, key)



from boa.blockchain.vm.Neo.Runtime import Notify

from .stuff.things import MoreAwesome, Awesome


def Main(a):

    # create an instance
    moar = MoreAwesome()

    awe = Awesome()

    count = moar.intval + 3

    count += awe.mycount

    # This does not work right now?
    # count2 = moar.intval + awe.mycount

    j = moar.multiply_nums(count, 3)

    Notify(j)

    thing = moar.do_something_with_awesome(awe)

    Notify(thing)

    mm = moar.make_awesome()

    Notify(mm)

    # this works
    r = moar.multiply_nums(awe.mycount, 4)

    # this doesnt work
    # r = moar.multiply_nums(awe.mycount, awe.mycount)

    r = moar.multiply_nums(5, awe.mycount)

    Notify(r)

    return count




class Awesome():
    """
    A very basic object
    """

    # it has some attributes
    myname = 'NEO'
    mycount = 25

    # and a method

    def awesome_name(self):
        """
        this demonstrates that the object can have te concept of ``self``

        Returns:
            return the value of the objects ``myname`` attribute
        """
        return self.myname


class MoreAwesome():
    """
    A little more complicated object
    """

    # it has some attributes
    thename = 'CoZ'
    intval = 32
    awesome = 42

    def mmmm(self, arg):

        print("hello1")
        return self.add_thirteen_to_this(arg)

    def add_thirteen_to_this(self, arg):
        """
        Demonstrate accepting an argument besides just ``self``
        Args:
            arg: any type

        Returns:
            return the value of the argument plus the value of ``self`` intval property
        """

        m = arg + self.intval

        return m

    def multiply_nums(self, a, b):
        """
        Demonstrates it can take more than 1 parameter
        Args:
            a: int: number to be multiplied
            b: int: number to be multiplied

        Returns:
            int: result of operation
        """
        return a * b

    def multiply_by_awesome(self, a):
        """
        Demonstrates accepting an argument and multiplying it by an attribute of ``self``
        Args:
            a: an int

        Returns:
            result of multiplication
        """
        return self.awesome * a

    def do_something_with_awesome(self, awe: Awesome):
        """
        Demonstrates that you can take another class instance as an argument
        and access a property of it
        Args:
            awe: (Awesome) an instance of the Awesome class

        Returns:
            the value of ``awe.mycount`` + 3
        """
        return awe.mycount + 3

    def make_awesome(self):
        """
        Demonstrates that you can instantiate a class inside another class
        Returns:
            the value of the new Awesome instance's ``mycount`` attribute minus 30
        """
        awe = Awesome()

        return awe.mycount - 30

    def instantiate_awesome(self, count_to_use) -> Awesome:

        awe = Awesome()
        awe.mycount = count_to_use

        return awe

    def accept_two_awesomes(self, awe1: Awesome, awe2: Awesome):

        total = awe1.mycount
        total += awe2.mycount
        return total



from boa.code.builtins import range


def Main():
    """

    :return:
    """
    count = 0

    r = range(0, 4)

    for i in range(0, 4):
        count += wah(i, count)
#        h = wah()
#        print(count)

    return count


def wah(i, count):

    return count + 1 * i



"""
NEP5 Token
===================================


.. moduleauthor:: Thomas Saunders <tom@cityofzion.io>

This file, when compiled to .avm format, would comply with the current NEP5 token standard on the NEO blockchain

Token standard is available in proposal form here:
`NEP5 Token Standard Proposal <https://github.com/neo-project/proposals/blob/master/nep-5.mediawiki>`_

Compilation can be achieved as such

>>> from boa.compiler import Compiler
>>> Compiler.load_and_save('./boa/tests/src/NEP5.py')

Or, from within the neo-python shell
``build path/to/NEP5.py test 0710 05 True name []``


Below is the current implementation in Python


"""

from boa.blockchain.vm.Neo.Runtime import Log, GetTrigger, CheckWitness
from boa.blockchain.vm.Neo.Action import RegisterAction
from boa.blockchain.vm.Neo.TriggerType import Application, Verification
from boa.blockchain.vm.Neo.Storage import GetContext, Get, Put, Delete
from boa.code.builtins import concat

# -------------------------------------------
# TOKEN SETTINGS
# -------------------------------------------

TOKEN_NAME = 'NEP5 Standard'
# Name of the Token

SYMBOL = 'NEP5'
# Symbol of the Token

OWNER = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
# Script hash of the contract owner

DECIMALS = 8
# Number of decimal places

TOTAL_SUPPLY = 123456789
# Total Supply of tokens in the system


# -------------------------------------------
# Events
# -------------------------------------------

DispatchTransferEvent = RegisterAction('transfer', 'from', 'to', 'amount')

DispatchApproveEvent = RegisterAction('approval', 'owner', 'spender', 'value')


def Main(operation, args):
    """
    This is the main entry point for the Smart Contract

    :param operation: the operation to be performed ( eg `balanceOf`, `transfer`, etc)
    :type operation: str
    :param args: a list of arguments ( which may be empty, but not absent )
    :type args: list
    :return: indicating the successful execution of the smart contract
    :rtype: bool
    """

    # The trigger determines whether this smart contract is being
    # run in 'verification' mode or 'application'

    trigger = GetTrigger()

    # 'Verification' mode is used when trying to spend assets ( eg NEO, Gas)
    # on behalf of this contract's address
    if trigger == Verification():

        # if the script that sent this is the owner
        # we allow the spend
        is_owner = CheckWitness(OWNER)

        if is_owner:

            return True

        return False

    # 'Application' mode is the main body of the smart contract
    elif trigger == Application():

        if operation == 'name':
            n = TOKEN_NAME
            return n

        elif operation == 'decimals':
            d = DECIMALS
            return d

        elif operation == 'symbol':
            sym = SYMBOL
            return sym

        elif operation == 'totalSupply':
            supply = TOTAL_SUPPLY
            return supply

        elif operation == 'balanceOf':
            if len(args) == 1:
                account = args[0]
                balance = BalanceOf(account)
                return balance
            return 0

        elif operation == 'transfer':
            if len(args) == 3:
                t_from = args[0]
                t_to = args[1]
                t_amount = args[2]
                transfer = DoTransfer(t_from, t_to, t_amount)
                return transfer
            else:
                return False

        elif operation == 'transferFrom':
            if len(args) == 3:
                t_from = args[0]
                t_to = args[1]
                t_amount = args[2]
                transfer = DoTransferFrom(t_from, t_to, t_amount)
                return transfer
            return False

        elif operation == 'approve':
            if len(args) == 3:
                t_owner = args[0]
                t_spender = args[1]
                t_amount = args[2]
                approve = DoApprove(t_owner, t_spender, t_amount)
                return approve
            return False

        elif operation == 'allowance':
            if len(args) == 2:
                t_owner = args[0]
                t_spender = args[1]
                amount = GetAllowance(t_owner, t_spender)
                return amount
            return False

        result = 'unknown operation'

        return result

    return False


def DoTransfer(t_from, t_to, amount):
    """
    Method to transfer NEP5 tokens of a specified amount from one account to another

    :param t_from: the address to transfer from
    :type t_from: bytearray
    :param t_to: the address to transfer to
    :type t_to: bytearray
    :param amount: the amount of NEP5 tokens to transfer
    :type amount: int

    :return: whether the transfer was successful
    :rtype: bool

    """
    if amount <= 0:
        Log("Cannot transfer negative amount")
        return False

    from_is_sender = CheckWitness(t_from)

    if not from_is_sender:
        Log("Not owner of funds to be transferred")
        return False

    if t_from == t_to:
        Log("Sending funds to self")
        return True

    context = GetContext()

    from_val = Get(context, t_from)

    if from_val < amount:
        Log("Insufficient funds to transfer")
        return False

    if from_val == amount:
        Delete(context, t_from)

    else:
        difference = from_val - amount
        Put(context, t_from, difference)

    to_value = Get(context, t_to)

    to_total = to_value + amount

    Put(context, t_to, to_total)

    DispatchTransferEvent(t_from, t_to, amount)

    return True


def DoTransferFrom(t_from, t_to, amount):
    """
    Method to transfer NEP5 tokens of a specified amount from one account to another

    :param t_from: the address to transfer from
    :type t_from: bytearray
    :param t_to: the address to transfer to
    :type t_to: bytearray
    :param amount: the amount of NEP5 tokens to transfer
    :type amount: int

    :return: whether the transfer was successful
    :rtype: bool

    """
    if amount <= 0:
        return False

    context = GetContext()

    allowance_key = concat(t_from, t_to)

    available_to_to_addr = Get(context, allowance_key)

    if available_to_to_addr < amount:
        Log("Insufficient funds approved")
        return False

    from_balance = Get(context, t_from)

    if from_balance < amount:
        Log("Insufficient tokens in from balance")
        return False

    to_balance = Get(context, t_to)

    # calculate the new balances
    new_from_balance = from_balance - amount
    new_to_balance = to_balance + amount
    new_allowance = available_to_to_addr - amount

    # persist the new balances
    Put(context, allowance_key, new_allowance)
    Put(context, t_to, new_to_balance)
    Put(context, t_from, new_from_balance)

    Log("transfer complete")

    # dispatch transfer event
    DispatchTransferEvent(t_from, t_to, amount)

    return True


def DoApprove(t_owner, t_spender, amount):
    """

    Method by which the owner of an address can approve another address
    ( the spender ) to spend an amount

    :param t_owner: Owner of tokens
    :type t_owner: bytearray
    :param t_spender: Requestor of tokens
    :type t_spender: bytearray
    :param amount: Amount requested to be spent by Requestor on behalf of owner
    :type amount: bytearray

    :return: success of the operation
    :rtype: bool

    """

    owner_is_sender = CheckWitness(t_owner)

    if not owner_is_sender:
        Log("Incorrect permission")
        return False

    context = GetContext()

    from_balance = Get(context, t_owner)

    # cannot approve an amount that is
    # currently greater than the from balance
    if from_balance >= amount:

        approval_key = concat(t_owner, t_spender)

        current_approved_balance = Get(context, approval_key)

        new_approved_balance = current_approved_balance + amount

        Put(context, approval_key, new_approved_balance)

        Log("Approved")

        DispatchApproveEvent(t_owner, t_spender, amount)

        return True

    return False


def GetAllowance(t_owner, t_spender):
    """
    Gets the amount of tokens that a spender is allowed to spend
    from the owners' account.

    :param t_owner: Owner of tokens
    :type t_owner: bytearray
    :param t_spender: Requestor of tokens
    :type t_spender: bytearray

    :return: Amount allowed to be spent by Requestor on behalf of owner
    :rtype: int

    """

    context = GetContext()

    allowance_key = concat(t_owner, t_spender)

    amount = Get(context, allowance_key)

    return amount


def BalanceOf(account):
    """
    Method to return the current balance of an address

    :param account: the account address to retrieve the balance for
    :type account: bytearray

    :return: the current balance of an address
    :rtype: int

    """
    context = GetContext()

    balance = Get(context, account)

    return balance



from boa.blockchain.vm.Neo.Runtime import Notify

from .stuff.things import Awesome


def Main(a):

    # create an instance
    m = Awesome()

    q = 1
    name = m.awesome_name()

    Notify(name)

    return q  # should be 2



from byteplay3 import Code, Opcode
from boa.code.method import Method
from boa.code.pytoken import PyToken

from boa.code import pyop
import importlib
import binascii
import sys
import os

from byteplay3 import Code, SetLinenoType, Label
from boa.code import pyop

from boa.code.line import Line
from boa.code.method import Method

from boa.blockchain.vm import VMOp

from collections import OrderedDict

import pdb


class Item():
    """

    """
    items = None

    def __init__(self, item_list):
        self.items = item_list

    def is_valid(self):
        """

        :return:
        """
        return True


class Definition(Item):

    value = None
    attr = None

    operations = None
    block = None

    is_method_call = False
    fn_call = None

    def __init__(self, item_list):
        super(Definition, self).__init__(item_list)

        # this is a simple definition like a = 3
        if len(self.items) == 3:
            self.value = PyToken(self.items[1], 1, args=self.items[1][1])
            self.attr = PyToken(self.items[2], 1, args=self.items[2][1])
        # a method based definition linke ctx = GetContext
        elif len(self.items) == 4:
            self.is_method_call = True
            self.attr = PyToken(self.items[-1], 1, args=self.items[-1][1])
            self.value = PyToken(Opcode(pyop.LOAD_CONST), 1, args=7)
            self.convert_class_call_to_block()

        elif len(self.items) == 5:
            if self.items[-1][0] == pyop.RETURN_VALUE:
                self.items = self.items[:3]
                self.value = PyToken(self.items[1], 1, args=self.items[1][1])
                self.attr = PyToken(self.items[2], 1, args=self.items[2][1])

        else:

            #            self.is_method_call=True
            self.attr = PyToken(self.items[-1], 1, args=self.items[-1][1])
#            print("SOMETHING ELSE!")
#            pdb.set_trace()

    def convert_class_call_to_block(self):

        from boa.code.block import Block

        opitems = self.items[1:]
        current_line_num = 1
        blockitems = []
        for i, (op, arg) in enumerate(opitems):

            token = PyToken(op, current_line_num, i, arg)
            blockitems.append(token)

        self.block = Block(blockitems)
        self.block.preprocess_method_calls(None)
        self.fn_call = self.block.oplist[0]


class Action(Item):

    """

    """
    event_name = None
    event_args = None

    method_name = None

    def __init__(self, item_list):
        super(Action, self).__init__(item_list)

        arguments = []

        for i, (key, value) in enumerate(self.items.items):
            if key == pyop.LOAD_CONST:
                arguments.append(value)
            elif key == pyop.STORE_NAME:
                self.method_name = value

        self.event_name = arguments[0]

        self.event_args = arguments


class SmartContractAppCall(Item):

    """

    """
    script_hash = None
    script_args = None

    method_name = None

    def __init__(self, item_list):
        super(SmartContractAppCall, self).__init__(item_list)

        arguments = []

        for i, (key, value) in enumerate(self.items.items):
            if key == pyop.LOAD_CONST:
                arguments.append(value)
            elif key == pyop.STORE_NAME:
                self.method_name = value

        self.script_hash = arguments[0]

        self.script_args = arguments

        if type(self.script_hash) is str:
            if len(self.script_hash) != 40:
                raise Exception(
                    "Invalid script hash! length of string must be 40")
        elif type(self.script_hash) in [bytes, bytearray]:
            if len(self.script_hash) != 20:
                raise Exception(
                    "Invalid Script hash, length in bytes must be 20")
        else:
            raise Exception(
                "Invalid script hash type.  must be string, bytes, or bytearray")

    @property
    def script_hash_addr(self):
        """

        :return:
        """

        return SmartContractAppCall.ToScriptHashData(self.script_hash)

    @staticmethod
    def ToScriptHashData(item):
        """

        :return:
        """
        b_array = None
        if type(item) is str:
            bstring = item.encode('utf-8')
            b_array = bytearray(binascii.unhexlify(bstring))
        elif type(item) is bytearray:
            pass
        elif type(item) is bytes:
            b_array = bytearray(item)
        else:
            raise Exception("Invalid script hash")

        b_array.reverse()

        return bytes(b_array)


class Import(Item):

    """

    """
    NEO_SC_FRAMEWORK = 'neo.SmartContract.Framework.'

    module_path = None
    module_name = None

    imported_module = None

    is_system_module = None

    module_items_to_import = None

    file_path = None

    def __init__(self, item_list, current_file_path):
        super(Import, self).__init__(item_list)
        self.module_items_to_import = []

        self.file_path = os.path.dirname(os.path.abspath(current_file_path))

        sys.path.append(self.file_path)

        for i, (op, arg) in enumerate(self.items):
            if op == pyop.IMPORT_NAME:
                self.module_path = arg
            elif op == pyop.STORE_NAME:
                self.module_items_to_import.append(arg)
#                print("SETTING MODALu nAme: %s " % self.module_name)

            elif op == pyop.IMPORT_STAR:
                self.module_items_to_import = ['STAR']

        self.is_system_module = False

        self.build()

    def build(self):
        # here is where we will check imports
        """

        """
        from boa.code.module import Module

        module = importlib.import_module(self.module_path, self.module_path)

        filename = module.__file__

        if self.NEO_SC_FRAMEWORK in self.module_path:
            self.is_system_module = True

        self.imported_module = Module(filename,
                                      module_name=self.module_path,
                                      is_sys_module=self.is_system_module,
                                      items_to_import=self.module_items_to_import)

    def is_valid(self):
        """

        :return:
        """
        return True

    def __str__(self):
        return "%s.%s" % (self.module_path, self.module_name)


class Klass(Item):

    """

    """
    name = None

    parent_name = None

    parent = None

    methods = None

    bp = None

    module = None

    lines = None

    class_vars = None
    class_method_calls = None

    @property
    def total_fields(self):
        return len(self.class_vars)

    @property
    def class_var_names(self):
        names = []
        for var in self.class_vars:
            names.append(var.attr.args)
        return names

    def index_of_varname(self, name):
        names = self.class_var_names
        return names.index(name)

    @property
    def all_method_names(self):
        return [m.full_name for m in self.methods]

    def __init__(self, item_list, module):
        super(Klass, self).__init__(item_list)
        self.module = self.parent = module
        self.methods = []
        self.lines = []
        self.class_vars = []
        self.class_method_calls = []
        self.build()

    def build(self):
        """

        """

        for i, (op, arg) in enumerate(self.items):

            # if the item is a byteplay3 code object, it is a method
            if type(arg) is Code:
                self.bp = arg

            # load name is called  to gather the class parent
            if op == pyop.LOAD_NAME:
                self.parent_name = arg

            # this occurs to store the name of the class
            if op == pyop.STORE_NAME:
                self.name = arg

        self.split_lines()

        for lineset in self.lines:

            if lineset.is_docstring:
                pass
            elif lineset.is_constant:
                self.class_vars.append(Definition(lineset.items))
            elif lineset.is_module_method_call:
                self.class_vars.append(Definition(lineset.items))

            elif lineset.is_method:

                m = Method(lineset.code_object, self)

                self.methods.append(m)

    def is_valid(self):
        # here is where we check if the class extends something reasonable
        """

        :return:
        """
        return True

    def split_lines(self):
        """
        Split the list of lines in the module into a set of objects that can be interpreted.
        """

        lineitem = None

        for i, (op, arg) in enumerate(self.bp.code):

            if isinstance(op, SetLinenoType):
                if lineitem is not None:
                    self.lines.append(Line(lineitem))

                lineitem = []

            lineitem.append((op, arg))

        if len(lineitem):
            self.lines.append(Line(lineitem))

    def __str__(self):
        return self.name




from byteplay3 import Opcode, Label
from boa.code.pytoken import PyToken
from boa.code import pyop
from boa.code.vmtoken import NEO_SC_FRAMEWORK
import pdb


class Block():

    """

    """
    forloop_counter = 0

    localmethod_counter = 0

    oplist = None  # list

    _label = None  # list

    iterable_variable = None
    iterable_loopcounter = None
    iterable_looplength = None
    iterable_item_name = None

    slice_item_length = None

    has_dynamic_iterator = False

    local_func_name = None
    local_func_varname = None

    def __init__(self, operation_list):
        self.oplist = operation_list

        self.iterable_variable = None
        self.iterable_loopcounter = None
        self.iterable_looplength = None
        self.iterable_item_name = None

        self.has_dynamic_iterator = False

        self.slice_item_length = None

    def __str__(self):
        if self._label:
            return '[Block] %s          [label] %s' % (self.oplist, self._label)
        return '[Block]: %s' % self.oplist

    def set_label(self, label):
        """

        :param label:
        """
        self._label = label
        self.oplist[0].jump_label = label

    @property
    def line(self):
        """

        :return:
        """
        if len(self.oplist):
            token = self.oplist[0]
            return token.line_no
        return None

    @property
    def has_store_fast(self):
        for token in self.oplist:
            if token.py_op == pyop.STORE_FAST:
                return True
        return False

    @property
    def has_load_attr(self):
        """

        :return:
        """
        for token in self.oplist:
            if token.py_op == pyop.LOAD_ATTR and token.instance_type is None:
                return True
        return False

    @property
    def has_store_attr(self):
        """

        Returns:
            bool
        """
        for token in self.oplist:
            if token.py_op == pyop.STORE_ATTR:
                return True
        return False

    @property
    def has_make_function(self):
        """

        :return:
        """
        for token in self.oplist:
            if token.py_op == pyop.MAKE_FUNCTION:
                return True
        return False

    @property
    def has_slice(self):
        """

        :return:
        """
        for token in self.oplist:
            if token.py_op == pyop.BUILD_SLICE:
                return True

    @property
    def is_return(self):
        """

        :return:
        """
        if len(self.oplist):
            token = self.oplist[-1]
            if token.py_op == pyop.RETURN_VALUE:
                return True
        return False

    @property
    def is_iter(self):
        """

        :return:
        """
        has_get_iter = False
        for token in self.oplist:
            if token.py_op == pyop.GET_ITER:
                has_get_iter = True
            elif token.py_op == pyop.MAKE_FUNCTION:
                return False
        return has_get_iter

    @property
    def iterable_local_vars(self):
        """

        :return:
        """
        return [
            self.iterable_looplength,
            self.iterable_loopcounter,
            self.iterable_item_name,
        ]

    @property
    def has_unprocessed_method_calls(self):
        """

        :return:
        """
        for token in self.oplist:
            if token.py_op == pyop.CALL_FUNCTION and not token.func_processed:
                return True
        return False

    @property
    def has_unprocessed_array_sub(self):
        """

        :return:
        """
        for token in self.oplist:
            if token.py_op == pyop.STORE_SUBSCR and not token.array_processed:
                return True
        return False

    @property
    def has_unprocessed_array(self):
        """

        :return:
        """
        for token in self.oplist:
            if token.py_op == pyop.BUILD_LIST and not token.array_processed:
                return True
        return False

    def preprocess_store_attr(self, method):

        for index, token in enumerate(self.oplist):
            if token.py_op == pyop.STORE_ATTR:
                to_store_into = self.oplist[index - 1].args

                if to_store_into == 'self':
                    ivar_type = method.parent
                    token.instance_type = ivar_type
                else:
                    try:
                        ivar_type = method.instance_vars[to_store_into]
                        token.instance_type = ivar_type
                        token.instance_name = to_store_into
                    except Exception as e:
                        print("Couldnt load instance variable: %s %s " % (to_store_into, e))

    def preprocess_load_attr(self, method):
        """

        :param method:
        """

        while self.has_load_attr:

            to_rep = {}
            to_del = []
            really_to_del = []
            for index, token in enumerate(self.oplist):

                index_to_rep = -1
                new_call = None

                if token.py_op == pyop.LOAD_ATTR:
                    from boa.code.items import Klass

                    to_load_from = self.oplist[index - 1]
                    varname = to_load_from.args
                    ivar_type = None
                    is_func_call = True
                    do_nothing = False
                    if varname in method.instance_vars.keys():
                        ivar_type = method.instance_vars[varname]

                        test_sysmodule = '%s.Get%s' % (ivar_type.module.module_path, token.args)

                        if NEO_SC_FRAMEWORK in test_sysmodule:
                            what_to_load = 'Get%s' % token.args

                        else:
                            what_to_load = '%s.%s' % (ivar_type.name, token.args)
                            token.instance_type = ivar_type

                            # if this is a class variable lookup, do this
                            if token.args in ivar_type.class_var_names:
                                is_func_call = False
                                what_to_load = token.args

                            # otherwise, it is a method call on an object, we don't want to do anything
                            else:
                                token.func_params = [self.oplist[index - 1]]
                                do_nothing = True

                    elif varname == 'self' and type(method.parent) is Klass:
                        ivar_type = method.parent

                        what_to_load = '%s.%s' % (str(ivar_type), token.args)
                        if token.args in ivar_type.class_var_names:
                            is_func_call = False
                            what_to_load = token.args

                    else:
                        what_to_load = 'Get%s' % token.args

                    if not do_nothing:
                        if is_func_call:
                            call_func = PyToken(
                                Opcode(pyop.CALL_FUNCTION), lineno=self.line, index=-1, args=what_to_load)
                            call_func.instance_type = ivar_type
                            call_func.func_processed = True
                            call_func.func_name = what_to_load
                            call_func.func_params = [self.oplist[index - 1]]
                            index_to_rep = index
                            new_call = call_func

                        else:
                            new_call = PyToken(Opcode(pyop.LOAD_CLASS_ATTR), lineno=self.line, args=what_to_load)
                            new_call.instance_type = ivar_type
                            new_call.instance_name = varname
                            new_call.func_processed = True
                            index_to_rep = index

                    if index_to_rep > 0:
                        to_rep[index_to_rep] = new_call
                        to_del.append(self.oplist[index_to_rep - 1])

            for key, val in to_rep.items():
                self.oplist[key] = val

            for item in to_del:
                #                print("WILL DELET: %s %s" % (item, vars(item)))
                if item in self.oplist:
                    self.oplist.remove(item)
                else:
                    pdb.set_trace()

#        print("oplist: %s " % [str(op) for op in self.oplist])
        # pdb.set_trace()

    def preprocess_load_class(self, method):
        print("PREPROCESS LOAD BUilD CLASS: %s %s " % (method, method.name))

    def preprocess_make_function(self, method):
        """

        :param method:
        """
        code_obj = self.oplist[0].args
        code_obj_name = self.oplist[1].args
        self.local_func_name = "%s_%s" % (
            code_obj_name, Block.localmethod_counter)
        Block.localmethod_counter += 1

        from boa.code.method import Method

        m = Method(code_object=code_obj, parent=method.parent,
                   make_func_name=self.local_func_name)
        method.parent.add_method(m)

        self.local_func_varname = self.oplist[-1].args

    def preprocess_slice(self):
        """
        this method processes slices of strings or byte arrays such as item[1:3]
        """
        index_to_remove = -1

        for index, token in enumerate(self.oplist):
            if token.py_op == pyop.BUILD_SLICE:
                # first, we want to take out the BINARY_SUBSC op, since we wont need it
                index_to_remove = index + 1

                # now we want to check the second item, for example item[2:4], we need to check 4
                # if you do item[2:], in python, normally the end is inferred
                # but we get None.
                # in that case, we need to convert None into the length of the item being sliced
                end_op = self.oplist[index - 1]

                if end_op.args is None:
                    # 0xffffff
                    # if you have a list greater than that length, well you'll have other problems than this
                    max_len = 16777215
                    end_op.args = max_len

        if index_to_remove > -1:
            del self.oplist[index_to_remove]

    def preprocess_iter(self):

        # in a better world this would be done in a more efficient way
        # for now this is kept to be as understandable as possible
        """

        """
        loopsetup = self.oplist[0]
        loopsetup.args = None
        loopsetup.jump_label = None

        # first we need to create a loop counter variable
        self.iterable_loopcounter = 'forloop_counter_%s' % Block.forloop_counter

        # load the value 0
        loopcounter_start_ld_const = PyToken(
            op=Opcode(pyop.LOAD_CONST), lineno=loopsetup.line_no, index=-1, args=0)
        # now store the loop counter
        loopcounter_store_fast = PyToken(op=Opcode(
            pyop.STORE_FAST), lineno=loopsetup.line_no, index=-1, args=self.iterable_loopcounter)

        # this loads the list that is going to be iterated over ( LOAD_FAST )
        # this will be removed... its added into the call get length token function params
        # unless this is a dynamic iteration, like for x in range(x,y)
        dynamic_iterable_items = []

        iterable_load = self.oplist[1]

        self.iterable_item_name = iterable_load.args

        if iterable_load.py_op == pyop.CALL_FUNCTION:

            self.has_dynamic_iterator = True

            self.iterable_item_name = 'forloop_dynamic_range_%s' % Block.forloop_counter

            dynamic_iterator_store_fast = PyToken(op=Opcode(pyop.STORE_FAST), lineno=loopsetup.line_no, index=-1,
                                                  args=self.iterable_item_name)

            # if we're calling a method in this for i in, like for i in range(x,y) then we need
            # to call the function
            dynamic_iterable_items = [
                iterable_load, dynamic_iterator_store_fast]

        # Now we need to get the length of that list, and store that as a local variable

        call_get_length_token = PyToken(
            op=Opcode(pyop.CALL_FUNCTION), lineno=loopsetup.line_no, args=1)
        call_get_length_token.func_params = [iterable_load]
        call_get_length_token.func_name = 'len'

        # now we need a variable name to store the length of the array
        self.iterable_looplength = 'forloop_length_%s' % Block.forloop_counter

        # now store the variable which is the output of the len(items) call
        looplength_store_op = PyToken(op=Opcode(
            pyop.STORE_FAST), lineno=loopsetup.line_no, index=-1, args=self.iterable_looplength)

        get_iter = self.oplist[2]
        for_iter = self.oplist[3]

        store_iterable_name = self.oplist[4]

        # set the iterable variable name ( for example, i ) so that the loop body can use it
        self.iterable_variable = store_iterable_name.args

        ld_loopcounter = PyToken(op=Opcode(
            pyop.LOAD_FAST), lineno=loopsetup.line_no, index=-1, args=self.iterable_loopcounter)

        ld_loop_length = PyToken(op=Opcode(
            pyop.LOAD_FAST), lineno=loopsetup.line_no, index=-1, args=self.iterable_looplength)

        new__compare_op = PyToken(
            op=Opcode(pyop.COMPARE_OP), lineno=loopsetup.line_no, index=-1, args='<')
        new__popjump_op = PyToken(op=Opcode(
            pyop.POP_JUMP_IF_FALSE), lineno=loopsetup.line_no, index=-1, args=for_iter.args)

        for_iter.args = None

        self.oplist = [
            loopsetup,  # SETUP_LOOP

            get_iter,  # GET_ITER, keep this in for now


            # the following 4 ops set up the iterator

            loopcounter_start_ld_const,  # LOAD_CONST 0
            loopcounter_store_fast,  # STORE_FAST forloopcounter_X

            # dynamic load loop stuff would go here

            call_get_length_token,  # CALL_FUNCTION 1

            looplength_store_op,  # STORE_FAST forloop_length_X


            # these last 5 ops controls the operation of the loop

            for_iter,  # tihs is the jump target for the end of the loop execution block

            ld_loopcounter,  # load in the loop counter LOAD_FAST forloopcounter_X

            ld_loop_length,  # load in the loop length LOAD_FAST forloop_length_X

            new__compare_op,  # COMPARE_OP <, this will compare foorloop_counter_X < forloop_length_X

            new__popjump_op  # POP_JUMP_IF_FALSE jumps to the loop exit when counter == length
        ]

        if len(dynamic_iterable_items):
            self.oplist.insert(4, dynamic_iterable_items[0])
            self.oplist.insert(5, dynamic_iterable_items[1])

        Block.forloop_counter += 1

    def process_iter_body(self, setup_block):
        """

        :param setup_block:
        """
        first_op = self.oplist[0]

        #
        # the following loads the iterated item into the block
        #

        # load the iterable collection
        ld_load_iterable = PyToken(op=Opcode(
            pyop.LOAD_FAST), lineno=first_op.line_no, index=-1, args=setup_block.iterable_item_name)

        # load the counter var
        ld_counter = PyToken(op=Opcode(pyop.LOAD_FAST), lineno=first_op.line_no,
                             index=-1, args=setup_block.iterable_loopcounter)

        # binary subscript of the iterable collection
        ld_subscript = PyToken(op=Opcode(pyop.BINARY_SUBSCR),
                               lineno=first_op.line_no, index=-1)

        # now store the iterated item
        st_iterable = PyToken(op=Opcode(
            pyop.STORE_FAST), lineno=first_op.line_no, index=-1, args=setup_block.iterable_variable)

        #
        # the following load the forloop counter and increments it
        #

        # load the counter var
        ld_counter_2 = PyToken(op=Opcode(
            pyop.LOAD_FAST), lineno=first_op.line_no, index=-1, args=setup_block.iterable_loopcounter)
        # load the constant 1
        increment_const = PyToken(
            op=Opcode(pyop.LOAD_CONST), lineno=first_op.line_no, index=-1, args=1)
        # add it to the counter
        increment_add = PyToken(
            op=Opcode(pyop.INPLACE_ADD), lineno=first_op.line_no, index=-1)
        # and store it again
        increment_store = PyToken(op=Opcode(
            pyop.STORE_FAST), lineno=first_op.line_no, index=-1, args=setup_block.iterable_loopcounter)

        self.oplist = [
            ld_load_iterable, ld_counter, ld_subscript, st_iterable,

            ld_counter_2, increment_const, increment_add, increment_store

        ] + self.oplist

    def lookup_return_types(self, orig_method):
        ivars = {}
        klass_type = None
        return_type = None
        for index, token in enumerate(self.oplist):
            if token.py_op == pyop.CALL_FUNCTION:
                param_count = token.args

                # why would param count be 256 when calling w/ kwargs?
                # when keyword args are sent, the param count is 256 * num paramms?
                if param_count % 256 == 0:
                    param_count = 2 * int(param_count / 256)

#                params = self.oplist[index - param_count:index]

                call_method_op = self.oplist[index - param_count - 1]
                call_method_name = call_method_op.args

                for method in orig_method.module.methods:
                    if method.name == call_method_name and method.return_type is not None:
                        klass_type = method.return_type

            if token.py_op == pyop.STORE_FAST and klass_type is not None:
                ivars[token.args] = klass_type
                klass_type = None

        return ivars

    def preprocess_method_calls(self, orig_method):
        """

        :param orig_method:
        """

        ivars = {}

        alreadythere = False

        while self.has_unprocessed_method_calls:
            start_index_change = None
            end_index_change = None
            changed_items = None

            klass = None

            for index, token in enumerate(self.oplist):

                if token.py_op == pyop.CALL_FUNCTION and not token.func_processed:

                    token.func_processed = True

                    param_count = token.args

                    # why would param count be 256 when calling w/ kwargs?
                    # when keyword args are sent, the param count is 256 * num paramms?
                    if param_count % 256 == 0:
                        param_count = 2 * int(param_count / 256)

                    params = self.oplist[index - param_count:index]

                    call_method_op = self.oplist[index - param_count - 1]
                    call_method_type = call_method_op.py_op
                    call_method_name = call_method_op.args

                    if call_method_op.instance_type:

                        # the call_method_op has a referecnce to the instance being called
                        # as the only item in the func_params ( this is python's self object )
                        # we will create a new param array by adding that one to the beginning of the method params

                        #                        print("PARAMS: %s "% params)
                        #                        print("COLL METHEHOD: %s "% call_method_op)
                        #                        print("call method name %s " % call_method_op.args)
                        #                        print("call method op func params: %s "% call_method_op.func_params)

                        #                        if call_method_name == 'owner':
                        #                            pdb.set_trace()

                        params = call_method_op.func_params + params

                        token.args = len(params)
                        param_count = token.args
                        if call_method_op.instance_type.name in call_method_op.args:
                            call_method_name = call_method_op.args
                            alreadythere = True

                        else:
                            call_method_name = "%s.%s" % (call_method_op.instance_type.name, call_method_op.args)

                    # we need to check if this is a method
                    # that is local to this block's method
                    if orig_method:
                        for key, value in orig_method.local_methods.items():
                            if key == call_method_name:
                                call_method_name = value

                    token.func_name = call_method_name
                    token.func_type = call_method_type

                    # check to see if this method call creates an instance of another object
                    if orig_method:
                        klass = orig_method.lookup_type(token.func_name)

                    if klass:
                        ivar_iname = self.oplist[2].args
                        ivars[ivar_iname] = klass

                    # if this method is the target of a jump
                    # or if one of its parameters is the target of a jump
                    # we need to catch that and use that jump label
                    # otherwise bad things
                    if call_method_op.jump_label is not None:
                        if len(params) > 0:
                            params[0].jump_label = call_method_op.jump_label
                        else:
                            token.jump_label = call_method_op.jump_label

                    token.func_params = params
                    changed_items = [token]

                    start_index_change = index - param_count - 1
                    end_index_change = index

            if start_index_change is not None and end_index_change is not None:
                tstart = self.oplist[0:start_index_change]
                tend = self.oplist[end_index_change + 1:]
                self.oplist = tstart + changed_items + tend

        if alreadythere:
            #            pdb.set_trace()

            if self.oplist[-1].py_op == pyop.STORE_FAST:
                print("Trimming load self method")
                self.oplist = self.oplist[-2:]

        return ivars

    def preprocess_array_subs(self):
        """

        """
        while self.has_unprocessed_array_sub:
            start_index_change = None
            end_index_change = None
            changed_items = None

            for index, token in enumerate(self.oplist):
                if token.py_op == pyop.STORE_SUBSCR and not token.array_processed:
                    token.array_processed = True
                    start_index_change = index - 3
                    end_index_change = index

                    item_to_sub = self.oplist[index - 3].args
                    array_to_sub = self.oplist[index - 2].args
                    index_to_sub_at = self.oplist[index - 1].args
                    changed_items = []

                    # load the array to set the item into
                    ld_op = PyToken(Opcode(pyop.LOAD_FAST),
                                    token.line_no, args=array_to_sub)
                    changed_items.append(ld_op)

                    # create the setitem op
                    settoken = PyToken(Opcode(
                        pyop.SETITEM), token.line_no, args=index_to_sub_at, array_item=item_to_sub)
                    changed_items.append(settoken)

            if start_index_change is not None and end_index_change is not None:
                tstart = self.oplist[0:start_index_change]
                tend = self.oplist[end_index_change + 2:]
                self.oplist = tstart + changed_items + tend

    def preprocess_arrays(self):
        """

        """
        while self.has_unprocessed_array:

            blist_start_index = None
            blist_end_index = None
            array_items = []
            for index, token in enumerate(self.oplist):
                if token.py_op == pyop.BUILD_LIST and not token.array_processed and blist_start_index is None:

                    num_list_items = token.args
                    blist_start_index = index - num_list_items
                    blist_end_index = index
                    array_items = self.oplist[index -
                                              num_list_items:num_list_items]
                    array_items.reverse()

                    token.array_processed = True

            if blist_start_index is not None:
                self.oplist = self.oplist[0:blist_start_index] + \
                    array_items + self.oplist[blist_end_index:]

    def mark_as_end(self):
        """

        """
        tstart = self.oplist[:-1]
        tend = self.oplist[-1:]

        newitems = [PyToken(Opcode(pyop.NOP), self.line),
                    #                    PyToken(pyop.DROP_BODY, self.line),
                    PyToken(Opcode(pyop.FROMALTSTACK), self.line),
                    PyToken(Opcode(pyop.DROP), self.line)]

        self.oplist = tstart + newitems + tend



from boa.blockchain.vm.Neo.Runtime import Notify

from .stuff.things import MoreAwesome


def Main(a):

    # create an instance
    m = MoreAwesome()

    q = 2

    m.intval = 13

    result = m.add_thirteen_to_this(q)

    Notify(result)  # should be fifteen

    mult = m.multiply_nums(3, 7)

    Notify(mult)

    atimes42 = m.multiply_by_awesome(2)

    Notify(atimes42)

    return result  # should be 15



from boa.blockchain.vm.Neo.Runtime import Notify

from .stuff.things import Awesome, MoreAwesome


def Main(a):

    # create an instance
    m1 = Awesome()
    m1.mycount = a

    m2 = Awesome()

    mm = MoreAwesome()

#       this works ok
    total = mm.accept_two_awesomes(m1, m2)
    Notify(total)

    # testing whether you can use more than one set of `LOAD_ATTR` in one operation
    total = m1.mycount + m2.mycount + m2.mycount
    total2 = 3 + m1.mycount
    total4 = m1.mycount + 3

    return total  # should be 53



from boa.blockchain.vm.Neo.Runtime import Notify

from .stuff.things import Awesome, MoreAwesome


def Main(a):

    # this should mess things up
    k = Awesome()

    m = what()  # type:MoreAwesome

    collection = [k, m, 5]

    k2 = collection[0]  # type:Awesome

    k2count = k2.mycount

    m2 = collection[1]  # type:MoreAwesome

    k2count += m2.intval

    j2 = collection[2]  # type:int

    Notify(k2count)

    Notify(j2)

    return 1


def what():

    q = MoreAwesome()

    return q



from boa.blockchain.vm.Neo.Runtime import Notify, GetTrigger, CheckWitness
from boa.blockchain.vm.Neo.Action import RegisterAction
from boa.blockchain.vm.Neo.TriggerType import Application, Verification
from boa.blockchain.vm.Neo.Blockchain import GetTransaction
from boa.blockchain.vm.Neo.TransactionType import InvocationTransaction
from boa.blockchain.vm.Neo.Transaction import *

from boa.blockchain.vm.System.ExecutionEngine import GetScriptContainer, GetExecutingScriptHash, GetCallingScriptHash, GetEntryScriptHash

from boa.blockchain.vm.Neo.Output import GetScriptHash, GetValue, GetAssetId
from boa.blockchain.vm.Neo.Input import GetHash, GetIndex
from boa.blockchain.vm.Neo.Storage import GetContext, Get, Put, Delete

from boa.code.builtins import range, concat, list, print_var

OWNER = b'\xaf\x12\xa8h{\x14\x94\x8b\xc4\xa0\x08\x12\x8aU\nci[\xc1\xa5'

NEO_ASSET_ID = b'\x9b|\xff\xda\xa6t\xbe\xae\x0f\x93\x0e\xbe`\x85\xaf\x90\x93\xe5\xfeV\xb3J\\"\x0c\xcd\xcfn\xfc3o\xc5'


onDeposit = RegisterAction('deposit', 'account', 'amount')

onWithdrawRequestApproved = RegisterAction('withdrawApproved', 'account', 'vin_requests')


def Main(operation, args):

    trigger = GetTrigger()

    if trigger == b'\x00':

        print("doing verification")

        # always allow owner to withdraw
        # this should be changed
        # to not allow the owner to withdraw from
        # vins that are earmarked for others
        if CheckWitness(OWNER):
            return True
        else:

            # if it is not the owner, we need to verify a few things
            can_widthraw = CanWithdraw()

            return can_widthraw

    elif trigger == b'\x10':

        if operation == 'deposit':

            deposit = DepositNeo()
            return deposit

        elif operation == 'withdrawalRequest':

            withdrawable_txs = VerifyWithdrawalRequest(args)
            return withdrawable_txs

        elif operation == 'getPending':
            account = args[0]
            withdrawable_txs = GetPendingWithdrawal(account)
            return withdrawable_txs

        elif operation == 'clearPending':
            account = args[0]
            result = DeletePendingWithdrawal(account)
            return result

        elif operation == 'balanceOf':
            account = args[0]
            result = BalanceOf(account)
            return result

        else:

            return 'unknown operation'

    return False


def CanWithdraw():

    return False


def DepositNeo():

    # get reference to the tx the invocation is in
    tx = GetScriptContainer()

    references = tx.References

    if len(references) < 1:
        print("no neo attached")
        return False

    # we need to determine who sent the tx
    reference = references[0]
    sender = GetScriptHash(reference)

    # this will keep track of how much neo was deposited
    value = 0

    # this is the contract's address
    receiver = GetExecutingScriptHash()

    # go through all the outputs
    # and check if the receiver is the contract
    # if that is true, we add the value of the output to the
    # total sum that was deposited

    for output in tx.Outputs:
        shash = GetScriptHash(output)
        if shash == receiver:

            output_asset_id = GetAssetId(output)

            if output_asset_id == NEO_ASSET_ID:

                output_val = GetValue(output)
                value = value + output_val

    if value > 0:

        # now we know value was deposited from the sender

        # check the current balance of the sender
        # add it to the deposit amount, and save
        context = GetContext()
        current_balance = Get(context, sender)
        new_balance = current_balance + value
        Put(context, sender, new_balance)

        # send deposit event
        onDeposit(sender, value)

        return True

    return False


def VerifyWithdrawalRequest(args):

    account = args[0]

    current_balance = BalanceOf(account)

    if current_balance == 0:
        print("No Current balance")
        return False

    if CheckHasPendingWithdrawal(account):
        print("already a pending withdrawal")
        return False

    arglen = len(args)

    txids_len = arglen / 2

    vin_requests = list(length=txids_len)

    # so, this is a bit involved, so bear with me here
    # lets assume a person has a balance that the contract owes them
    # we need to allow them to put a hold on some tx vins that they can 'claim'
    # after they have put the hold on them, they can withdraw and in the
    # verification portion of the contract it will allow them

    # but they may need to put a hold on tx vins that have a value
    # more than they are owed.  So we want to allow them to put a hold
    # on x number of txids, where the amount contained in x-1 vins is less than the amount they are owed
    # and the amount in x txids is greater than or equal to the amount they are owed
    # but only by one

    hold_amount = 0
    okcount = 0

    for i in range(0, txids_len):

        j = i * 2

        request_vin_tx = args[j + 1]
        request_vin_index = args[j + 2]

        vin = LookupVIN(request_vin_tx, request_vin_index)

        if vin and vin[1] == NEO_ASSET_ID:

            actual_amount = vin[0]

            # if the amount being held is less than the current balance
            # it is ok, we allow the vin request
            # this allows the hold amount to go over
            # but if it is too much then invalidate
            # the entire request

            if hold_amount < current_balance:
                hold_amount += actual_amount

                vin_request = [request_vin_tx, request_vin_index]
                vin_requests[i] = vin_request
                okcount += 1

    if okcount != txids_len:
        print("Invalid txid request")
        return False

    Notify(hold_amount)
    Notify(current_balance)

    # now that we've checked the requestor isn't requesting too much
    # we will check that the requestor can put a hold on the vins
    # we make sure they can put a hold on all the vins
    # before actually placing the hold on the vin
    for vin in vin_requests:

        vin_tx = vin[0]
        vin_index = vin[1]

        has_hold = HasVINHold(vin_tx, vin_index)

        if has_hold:
            print("Cannot request withdrawal, vin(s) already have hold")
            return False

    # now that we're sure we can put holds on all the requested vins
    # we will put holds on the requested vins

    output = None

    for vin in vin_requests:
        vin_tx = vin[0]
        vin_index = vin[1]

        hold_req = PlaceVINHold(account, vin_tx, vin_index)

        # now we assemble a bytearray
        # to save to storage
        vin = concat(vin_index, vin_tx)

        if len(output) > 1:
            output = concat(output, vin)
        else:
            output = vin

    onWithdrawRequestApproved(account, vin_requests)

    m = SetPendingWithdrawal(account, output)

    return vin_requests


def BalanceOf(account):
    """
    Method to return the current balance of an address

    :param account: the account address to retrieve the balance for
    :type account: bytearray

    :return: the current balance of an address
    :rtype: int

    """
    context = GetContext()
    balance = Get(context, account)

    return balance


def CheckHasPendingWithdrawal(account):

    context = GetContext()

    account_has_pending = concat(account, 'pending')

    pending = Get(context, account_has_pending)

    return pending


def SetPendingWithdrawal(account, value):

    context = GetContext()

    account_has_pending = concat(account, 'pending')

    if value == 0:

        Delete(context, account_has_pending)

    else:

        Put(context, account_has_pending, value)

    return True


def GetPendingWithdrawal(account):

    context = GetContext()

    account_pending = concat(account, 'pending')

    result = Get(context, account_pending)

    return result


def DeletePendingWithdrawal(account):

    context = GetContext()
    account_pending = concat(account, 'pending')

    Delete(context, account_pending)

    return True


def HasVINHold(txid, index):

    context = GetContext()
    hold_id = concat(txid, index)

    item = Get(context, hold_id)

    return item


def PlaceVINHold(account, txid, index):

    context = GetContext()

    hold_id = concat(index, txid)

    Put(context, hold_id, account)

    return True


def GetTXInputs():

    tx = GetScriptContainer()

    inputs = tx.Inputs
    inputlen = len(inputs)

    results = list(length=inputlen)

    count = 0

    for input in tx.Inputs:

        txid = GetHash(input)
        index = GetIndex(input)
        item = [txid, index]
        results[count] = item
        count += 1

    return results


def LookupVIN(txid, index):

    tx = GetTransaction(txid)

    if tx:

        outputs = tx.Outputs

        output_len = len(outputs)

        if index < output_len:

            output = outputs[index]

            assetType = GetAssetId(output)
            assetAmount = GetValue(output)
            toret = [assetAmount, assetType]
            return toret

    print("could not lookup vin. TX or output not found")
    return False



from boa.blockchain.vm.Neo.Runtime import Notify
from .stuff.storage_api import StorageAPI


def Main(a):

    storage = StorageAPI()

    storage.putitem('hello', 9)

    helloval = storage.getitem('hello')

    Notify(helloval)

    storage.deleteitem('hello')

    hval2 = storage.getitem('hello')

    Notify(hval2)

    # put it again
    storage.putitem('hello', 6)

    m = storage.getitem('hello')

    storage.putitem('h1', 2)
    storage.putitem('h2', 5)

    # this messes up the stack
    # res = storage.getitem('h1') + storage.getitem('h2')

    h1 = storage.getitem('h1')

    # this is ok?
    res = h1 + storage.getitem('h2')

    return res



from boa.blockchain.vm.Neo.Runtime import Notify

from .stuff.things import Awesome


def Main(a):

    # create an instance
    m = Awesome()

    # passing an item to a method doesn't work
    q = second(m, 8, m)

    return q  # should be 2


def second(item: Awesome, num: int, item2: Awesome):

    j = item.mycount

    total = j + num + item2.mycount

    Notify(total)

    return total



from boa.blockchain.vm.Neo.Runtime import GetTrigger
from boa.blockchain.vm.Neo.TriggerType import Application, Verification
from boa.code.builtins import substr


def Main(bytes):

    trigger = GetTrigger()

    if trigger == Verification():
        print("Verification!")
        return False

    elif trigger == Application():
        print("Application!")
        i = 0
        l = len(bytes)
        while i < l:
            c = substr(bytes, i, 1)
            i = i + 1
            if c == 0:
                print("Null")
            else:
                print("Not null")
                return True



from boa.blockchain.vm.Neo.Runtime import Notify

from .stuff.things import Awesome, MoreAwesome


def Main(a):

    b = CreateAwesome()

    q = b.mycount

    m = MoreAwesome()

    j = m.instantiate_awesome(7)  # type:Awesome

    Notify(j.mycount)

    return q  # should be 25


def CreateAwesome() -> Awesome:

    a = Awesome()
    return a



from .stuff.things import Awesome, MoreAwesome


def Main(a):

    b = 10  # type:int

    k = Awesome()  # type:Awesome

    m = what()  # type:MoreAwesome

    j = 'stnahoeu'  # type:str

    q = b'\x00'  # type:bytes

    ba = bytearray(b'\x00\x01')  # type:bytearray

    return m.awesome


def what():

    q = MoreAwesome()

    return q



from boa.blockchain.vm.Neo.Runtime import Notify

from .stuff.things import Awesome, MoreAwesome


def Main(a):

    # create an instance
    m1 = Awesome()
    m1.mycount = a

    m2 = Awesome()
    m2.mycount = 4 * a

    mm = MoreAwesome()

    res = mm.multiply_nums(m1.mycount, m2.mycount)

    Notify(res)

    return res
